/*
 * Simple waveform generator
 * Parts:
 * AD9833 dev board. The cheap blue one with 7 pins.
 * SSD1306 I2C OLED (128x64)
 * Rotary Encoder with push switch
 * 3 momentary switches
 * Arduino UNO
 * 
 */


// SCL - A5
// SDA - A4

#include <U8x8lib.h>
U8X8_SSD1306_128X64_NONAME_HW_I2C u8x8(U8X8_PIN_NONE);

#include "MD_AD9833.h"
#define FSYNC_PIN 10
// SDATA - D11
// SCLK  - D13
MD_AD9833 ad9833 = MD_AD9833(FSYNC_PIN);

// PCINT18 = Pin D2
#define LEFT_BUTTON 2

// PCINT19 = Pin D3
#define ENCODER_BUTTON 3

// PCINT20 = Pin D4
#define ENCODER_CW 4

// PCINT21 = Pin D5
#define ENCODER_ANTI_CW 5

// PCINT22 = Pin D6
#define RIGHT_BUTTON 6

// PCINT23 = Pin D7
#define WAVE_TYPE_BUTTON 7

#define WAVE_TYPE_SINE 0
#define WAVE_TYPE_TRIANGLE 1
#define WAVE_TYPE_SQUARE 2
int selected_wave_type = WAVE_TYPE_SINE;

// True if something was changed in an interrupt handler
bool dirty = false;

// Currently selected digit
// 0 is the leftmost, 6 the rightmost, -1 none selected
const int max_selected_digit_idx = 6;
int selected_digit = 0;

// Flag to tell us if we're changing the signal generator output
bool set_output = false;

// Display from 1Hz to 999999
// Store each digit in an array to make it easy to draw the currently 
// selected digit in reverse font
int digits[max_selected_digit_idx + 1] = {0, 0, 0, 0, 0, 0, 1};

void setup() {
 
  pinMode(ENCODER_BUTTON, INPUT_PULLUP);
  pinMode(ENCODER_ANTI_CW, INPUT);
  pinMode(ENCODER_CW, INPUT);
  pinMode(LEFT_BUTTON, INPUT_PULLUP);
  pinMode(RIGHT_BUTTON, INPUT_PULLUP);
  pinMode(WAVE_TYPE_BUTTON, INPUT_PULLUP);

  PCICR =  0b00000100;   // PCIE2: Pin Change Interrupt Enable 2
  PCMSK2 = 0b11111100;   // Enable Pin Change Interrupt for D2 - D6

  u8x8.begin();
  u8x8.setFont(u8x8_font_chroma48medium8_r);

  draw_frequency();
  draw_wave_type();

  ad9833.begin();
  ad9833.reset(true);
  ad9833.setFrequency(MD_AD9833::CHAN_0, 0.1f);
  ad9833.setPhase(MD_AD9833::CHAN_0, 0.0f);
  ad9833.setActiveFrequency(MD_AD9833::CHAN_0);
  ad9833.setActivePhase(MD_AD9833::CHAN_0);
  ad9833.setMode(MD_AD9833::MODE_SINE);
  ad9833.reset();
}

void loop() {
  while (true) {

    if (dirty) {
      draw_frequency();
      draw_wave_type();
      dirty = false;
    }

    if (set_output) {
      set_output = false;
      update_generator();
    }

    delay(10);
  }
}

/*
 * The user pressed the button on the rotary encoder so set the frequency
 * and wave type
 */
void update_generator() {

  // show feedback
  long hz = number_from_digits();
  char buf[10];
  sprintf(buf, "%ld", hz);
  u8x8.drawString(0, 3, "Changing output");
  u8x8.drawString(0, 4, buf);

  ad9833.setFrequency(MD_AD9833::CHAN_0, hz);

  switch (selected_wave_type) {
    case WAVE_TYPE_SINE:
      ad9833.setMode(MD_AD9833::MODE_SINE);
      u8x8.drawString(0, 5, "Sine");
      break;
    case WAVE_TYPE_TRIANGLE:
      ad9833.setMode(MD_AD9833::MODE_TRIANGLE);
      u8x8.drawString(0, 5, "Triangle");
      break;
    case WAVE_TYPE_SQUARE:
      ad9833.setMode(MD_AD9833::MODE_SQUARE1);
      u8x8.drawString(0, 5, "Square");
      break;
    default:
      u8x8.drawString(0, 5, "Bad wave type");
      break;
  }
}

// Draw the freqency on the OLED with the selected digit
// in inverse font
void draw_frequency() {
  u8x8.clear();

  char buf[10];
  int i;
  for (i = 0; i <= max_selected_digit_idx; i++) {
    Serial.print(digits[i]); Serial.print(" ");
    u8x8.setInverseFont(i == selected_digit ? true : false);
    sprintf(buf, "%d", digits[i]);
    u8x8.drawString(i, 0, buf);
  }

  u8x8.setInverseFont(false);
  u8x8.drawString(i + 1, 0, "Hz");
}

static char *strWaveType = "??";
void draw_wave_type() {
  if (selected_wave_type == WAVE_TYPE_SINE) {
    strWaveType = "Sine";
  }
  else if (selected_wave_type == WAVE_TYPE_TRIANGLE) {
    strWaveType = "Triangle";
  }
  else if (selected_wave_type == WAVE_TYPE_SQUARE) {
    strWaveType = "Square";
  }

  u8x8.drawString(0, 1, strWaveType);
}

// We store the frequency as individual digits to make it easy to print it
// with the selected digit highlighted. This just returns the actual frequency
// as a long
long number_from_digits() {
  long num = 0;
  for (int i = 0; i <= max_selected_digit_idx; i++) {
    int exponent = max_selected_digit_idx - i;
    num += digits[i] * pow10(exponent);
  }

  return num;
}

// Was getting rounding errors when using the inbuilt pow() function
// I couldn't be bothered figuring out why so just using
// this for now
long pow10(int exponent) {
  long num = 1;
  for (int i = 0; i < exponent; i++) {
    num *= 10;
  }
  return num;
}


ISR (PCINT2_vect) {

  static uint16_t seqA, seqB;

  // If interrupt is triggered by the button
  if (!digitalRead(RIGHT_BUTTON)) {
    if (selected_digit == -1) {
      selected_digit = max_selected_digit_idx;
    }
    else if (selected_digit == max_selected_digit_idx) {
      selected_digit = 0;
    }
    else {
      selected_digit++;
    }
    dirty = true;
  }
  else if (!digitalRead(ENCODER_BUTTON)) {
    set_output = true;
  }
  else if (!digitalRead(LEFT_BUTTON)) {
    if (selected_digit == -1) {
      selected_digit = 0;
    }
    else if (selected_digit == 0) {
      selected_digit = max_selected_digit_idx;
    }
    else {
      selected_digit--;
    }
    dirty = true;
  }
  else if (!digitalRead(WAVE_TYPE_BUTTON)) {
    if (selected_wave_type >= 2) {
      selected_wave_type = 0;
    }
    else {
      selected_wave_type++;
    }
    dirty = true;
  }
  else
  {
    // interrupt generated by encoder dial

    // Read A and B signals
    boolean A_val = digitalRead(ENCODER_ANTI_CW);
    boolean B_val = digitalRead(ENCODER_CW);

    // Record the A and B signals in seperate sequences
    seqA <<= 1;
    seqA |= A_val;

    seqB <<= 1;
    seqB |= B_val;

    // Mask the MSB four bits
    seqA &= 0b00001111;
    seqB &= 0b00001111;

    // Compare the recorded sequence with the expected sequence
    int selected_digit_value = digits[selected_digit];
    if (seqA == 0b00001001 && seqB == 0b00000011) {
      if (selected_digit_value == 0) {
        selected_digit_value = 9;
      }
      else {
        selected_digit_value--;
      }
      dirty = true;
    }
    else if (seqA == 0b00000011 && seqB == 0b00001001) {
      if (selected_digit_value == 9) {
        selected_digit_value = 0;
      }
      else {
        selected_digit_value++;
      }
      dirty = true;
    }
    digits[selected_digit] = selected_digit_value;
  }
}
